@@ -1,9 +1,10 @@
-'''Data manager and advanced queries for Redis.'''
-from .exceptions import *
-from .backends import *
+'''Object data mapper and advanced query manager for non relational databases.'''
+from .utils.exceptions import *
+from .utils.conf import settings
 from .utils.version import get_version, stdnet_version
+from .backends import *
 
-VERSION = stdnet_version(0, 7, 0, 'final', 1)
+VERSION = stdnet_version(0, 8, 0, 'final', 1)
 
 __version__ = version = get_version(VERSION)
 __license__ = "BSD"
@@ -27,6 +28,4 @@ CLASSIFIERS = [
                'Topic :: Utilities',
                'Topic :: Database',
                'Topic :: Internet'
-               ]
-
-sphinxtogithub = True
\ No newline at end of file
+               ]
\ No newline at end of file
