@@ -1,30 +1,21 @@
 import logging
 from copy import copy
-from hashlib import sha1
-from collections import namedtuple
-import time
 from datetime import date, datetime
 from base64 import b64encode
 
 from stdnet import range_lookups
-from stdnet.exceptions import *
-from stdnet.utils import pickle, DefaultJSONEncoder,\
-                         DefaultJSONHook, timestamp2date, date2timestamp,\
-                         UnicodeMixin, to_string, is_string,\
-                         is_bytes_or_string, iteritems,\
-                         encoders, flat_to_nested, dict_flat_generator,\
-                         string_type
+from stdnet.utils import DefaultJSONEncoder, DefaultJSONHook, timestamp2date,\
+                         date2timestamp, UnicodeMixin, to_string, string_type,\
+                         encoders, flat_to_nested, dict_flat_generator
+from stdnet.utils.exceptions import *
 
 from . import related
 from .globals import get_model_from_hash, get_hash_from_model, JSPLITTER
 
-
-orderinginfo = namedtuple('orderinginfo','name field desc model nested, auto')
-
 logger = logging.getLogger('stdnet.odm')
 
 __all__ = ['Field',
-           'AutoField',
+           'AutoIdField',
            'AtomField',
            'IntegerField',
            'BooleanField',
@@ -46,15 +37,13 @@ NONE_EMPTY = (None,'')
 
 
 def field_value_error(f):
-
-    def _(self, value):
+    def _(self, value, *args, **kwargs):
         try:
-            return f(self, value)
+            return f(self, value, *args, **kwargs)
         except FieldValueError:
             raise
         except:
             raise FieldValueError('%s not valid for "%s"' % (value, self.name))
-
     _.__name__ = f.__name__
     _.__doc__ = f.__doc__
     return _
@@ -95,7 +84,7 @@ Each field is specified as a :class:`StdModel` class attribute.
     * :attr:`Field.unique` is also ``True``.
     * There can be only one in a model.
     * It's attribute name in the model must be **id**.
-    * If not specified a :class:`AutoField` will be added.
+    * If not specified a :class:`AutoIdField` will be added.
 
     Default ``False``.
 
@@ -151,18 +140,16 @@ Each field is specified as a :class:`StdModel` class attribute.
 
     Default ``False``.
 '''
-    default = None
+    _default = None
     type = None
     python_type = None
     index = True
-    ordered = False
     charset = None
     hidden = False
     internal_type = None
 
-    def __init__(self, unique=False, ordered=None, primary_key=False,
-                 required=True, index=None, hidden=None, as_cache=False,
-                 **extras):
+    def __init__(self, unique=False, primary_key=False, required=True,
+                 index=None, hidden=None, as_cache=False, **extras):
         self.primary_key = primary_key
         index = index if index is not None else self.index
         if primary_key:
@@ -170,6 +157,7 @@ Each field is specified as a :class:`StdModel` class attribute.
             self.required = True
             self.index = True
             self.as_cache = False
+            extras['default'] = None
         else:
             self.unique = unique
             self.required = required
@@ -180,15 +168,22 @@ Each field is specified as a :class:`StdModel` class attribute.
             self.unique = False
             self.index = False
         self.charset = extras.pop('charset',self.charset)
-        self.ordered = ordered if ordered is not None else self.ordered
         self.hidden = hidden if hidden is not None else self.hidden
         self.meta = None
         self.name = None
         self.model = None
-        self.default = extras.pop('default',self.default)
+        self._default = extras.pop('default', self._default)
         self.encoder = self.get_encoder(extras)
         self._handle_extras(**extras)
 
+    @property
+    def default(self):
+        return self._default
+    
+    @property
+    def class_field(self):
+        return False
+    
     def get_encoder(self, params):
         return None
 
@@ -202,7 +197,7 @@ Each field is specified as a :class:`StdModel` class attribute.
         return to_string('%s.%s' % (self.meta,self.name))
 
     def value_from_data(self, instance, data):
-        return data.pop(self.attname,None)
+        return data.pop(self.attname, None)
 
     def register_with_model(self, name, model):
         '''Called during the creation of a the :class:`StdModel`
@@ -221,6 +216,8 @@ function users should never call.'''
         meta.fields.append(self)
         if not self.primary_key:
             self.add_to_fields()
+        else:
+            model._meta.pk = self
 
     def add_to_fields(self):
         meta = self.model._meta
@@ -233,6 +230,8 @@ function users should never call.'''
         return self.name
 
     def get_cache_name(self):
+        '''name for the private attribute which contains a cached value
+for this field. Used only by realted fields.'''
         return '_%s_cache' % self.name
 
     def id(self, obj):
@@ -241,10 +240,11 @@ function users should never call.'''
 
     def get_default(self):
         "Returns the default value for this field."
-        if hasattr(self.default,'__call__'):
-            return self.default()
+        default = self._default
+        if hasattr(default, '__call__'):
+            return default()
         else:
-            return self.default
+            return default
 
     def __deepcopy__(self, memodict):
         '''Nothing to deepcopy here'''
@@ -259,15 +259,44 @@ function users should never call.'''
 
     def get_sorting(self, name, errorClass):
         raise errorClass('Cannot use nested sorting on field {0}'.format(self))
+    
+    def get_lookup(self, remaining, errorClass=ValueError):
+        '''called by the :class:`Query` method when it needs to build
+lookup on fields with additional nested fields. This is the case of
+:class:`ForeignKey` and :class:`JSONField`.
+
+:param remaining: the :ref:`double underscored` fields if this :class:`Field`
+:param errorClass: Optional exception class to use if the *remaining* field
+    is not valid.'''
+        if remaining:
+            raise errorClass('Cannot use nested lookup on field %s' % self)
+        return (self.attname, None)
 
     def todelete(self):
         return False
-
+    
     ############################################################################
-    ##    FIELD CONVERTERS
+    ##    FIELD VALUES
     ############################################################################
+    def get_value(self, instance, *bits):
+        '''Retrieve the value :class:`Field` from a
+:class:`StdModel` ``instance``.
 
-    def to_python(self, value):
+:param instance: The :class:`StdModel` ``instance`` invoking this function.
+:param bits: Additional information for nested fields which derives from
+    the :ref:`double underscore <tutorial-underscore>` notation.
+:return: the value of this :class:`Field` in the ``instance``. can raise
+    :class:`AttributeError`.
+    
+This method is used by the :meth:`StdModel.get_attr_value` method when
+retrieving values form a :class:`StdModel` instance.
+'''
+        if bits:
+            raise AttributeError
+        else:
+            return getattr(instance, self.attname)
+    
+    def to_python(self, value, backend=None):
         """Converts the input value into the expected Python
 data type, raising :class:`stdnet.FieldValueError` if the data
 can't be converted.
@@ -313,7 +342,7 @@ value with a specific data type. it can be of four different types:
 * floating point
 * symbol
 '''
-    def to_python(self, value):
+    def to_python(self, value, backend=None):
         if hasattr(value, '_meta'):
             return value.pkvalue()
         else:
@@ -321,18 +350,42 @@ value with a specific data type. it can be of four different types:
     json_serialize = to_python
 
 
+class AutoIdField(AtomField):
+    '''An :class:`AtomField` for primary keys which are automatically
+generated by the backend server.
+You usually won't need to use this directly; a ``primary_key`` field
+of this type, named ``id``, will automatically be added to your model
+if you don't specify otherwise.
+Check the :ref:`primary key tutorial <tutorial-primary-unique>` for
+further information on primary keys.'''
+    type = 'auto'
+    
+    def __init__(self, *args, **kwargs):
+        kwargs['primary_key'] = True
+        super(AutoIdField, self).__init__(*args, **kwargs)
+    
+    @field_value_error
+    def to_python(self, value, backend=None):
+        if hasattr(value, '_meta'):
+            return value.pkvalue()
+        elif backend:
+            return backend.auto_id_to_python(value)
+        else:
+            return value
+        
+
 class BooleanField(AtomField):
     '''A boolean :class:`AtomField`'''
     type = 'bool'
     internal_type = 'numeric'
     python_type = bool
-    default = False
+    _default = False
 
     def __init__(self, required = False, **kwargs):
         super(BooleanField,self).__init__(required = required,**kwargs)
 
     @field_value_error
-    def to_python(self, value):
+    def to_python(self, value, backend=None):
         if value in NONE_EMPTY:
             return self.get_default()
         else:
@@ -348,27 +401,16 @@ class IntegerField(AtomField):
     type = 'integer'
     internal_type = 'numeric'
     python_type = int
-    #default = 0
 
     @field_value_error
-    def to_python(self, value):
-        value = super(IntegerField,self).to_python(value)
+    def to_python(self, value, backend=None):
+        value = super(IntegerField, self).to_python(value, backend)
         if value in NONE_EMPTY:
             return self.get_default()
         else:
             return self.python_type(value)
 
 
-class AutoField(IntegerField):
-    '''An :class:`IntegerField` that automatically increments.
-You usually won't need to use this directly;
-a ``primary_key`` field  of this type, named ``id``,
-will automatically be added to your model
-if you don't specify otherwise.
-    '''
-    type = 'auto'
-
-
 class FloatField(IntegerField):
     '''An floating point :class:`AtomField`. By default
 its :attr:`Field.index` is set to ``False``.
@@ -385,11 +427,10 @@ a :class:`datetime.date` instance.'''
     type = 'date'
     internal_type = 'numeric'
     python_type = date
-    ordered = True
-    default = None
+    _default = None
 
     @field_value_error
-    def to_python(self, value):
+    def to_python(self, value, backend=None):
         if value not in NONE_EMPTY:
             if isinstance(value,date):
                 if isinstance(value,datetime):
@@ -420,7 +461,7 @@ a :class:`datetime.datetime` instance.'''
     index = False
 
     @field_value_error
-    def to_python(self, value):
+    def to_python(self, value, backend=None):
         if value not in NONE_EMPTY:
             if isinstance(value,date):
                 if not isinstance(value,datetime):
@@ -441,13 +482,13 @@ or other entities. They are indexes by default.'''
     python_type = string_type
     internal_type = 'text'
     charset = 'utf-8'
-    default = ''
+    _default = ''
 
     def get_encoder(self, params):
         return encoders.Default(self.charset)
 
     @field_value_error
-    def to_python(self, value):
+    def to_python(self, value, backend=None):
         value = super(SymbolField,self).to_python(value)
         if value is not None:
             return self.encoder.loads(value)
@@ -479,10 +520,10 @@ In python this is converted to `bytes`.'''
     type = 'bytes'
     internal_type = 'bytes'
     python_type = bytes
-    default = b''
+    _default = b''
 
     @field_value_error
-    def to_python(self, value):
+    def to_python(self, value, backend=None):
         if value is not None:
             return self.encoder.loads(value)
         else:
@@ -508,7 +549,7 @@ or :class:`JSONField` fields as more general alternatives.
           attribute is ``True``.
 '''
     type = 'object'
-    default = None
+    _default = None
 
     def serialize(self, value):
         if value is not None:
@@ -538,8 +579,10 @@ Behind the scenes, stdnet appends "_id" to the field name to create
 its field name in the back-end data-server. In the above example,
 the database field for the ``File`` model will have a ``folder_id`` field.
 
-It accepts **related_name** as extra argument. It is the name to use for
-the relation from the related object back to self.
+.. attribute:: related_name
+
+    Optional name to use for the relation from the related object
+    back to ``self``.
 '''
     type = 'related object'
     internal_type = 'numeric'
@@ -560,21 +603,19 @@ the relation from the related object back to self.
     def register_with_related_model(self):
         # add the RelatedManager proxy to the model holding the field
         setattr(self.model, self.name, self.proxy_class(self))
-        setattr(self.model, self.get_query_attname(),
-                related.LazyForeignQuery(self))
         related.load_relmodel(self, self._set_relmodel)
 
     def _set_relmodel(self, relmodel):
         self.relmodel = relmodel
         meta  = self.relmodel._meta
-        related_name = self.related_name or '%s_set' % self.model._meta.name
-        if related_name not in meta.related and related_name\
+        if not self.related_name:
+            self.related_name = '%s_%s_set' % (self.model._meta.name, self.name)
+        if self.related_name not in meta.related and self.related_name\
                                                  not in meta.dfields:
-            self.related_name = related_name
             self._register_with_related_model()
         else:
             raise FieldError('Duplicated related name "{0}"\
- in model "{1}" and field {2}'.format(related_name,meta,self))
+ in model "{1}" and field {2}'.format(self.related_name, meta, self))
 
     def _register_with_related_model(self):
         manager = self.related_manager_class(self)
@@ -584,9 +625,10 @@ the relation from the related object back to self.
 
     def get_attname(self):
         return '%s_id' % self.name
-
-    def get_query_attname(self):
-        return '%s_query' % self.name
+        
+    def get_value(self, instance, *bits):
+        related = getattr(instance, self.name)
+        return related.get_attr_value(JSPLITTER.join(bits)) if bits else related
 
     def register_with_model(self, name, model):
         super(ForeignKey,self).register_with_model(name, model)
@@ -600,20 +642,40 @@ the relation from the related object back to self.
             raise FieldValueError('cannot evaluate score of {0}'.format(value))
 
     @field_value_error
-    def to_python(self, value):
-        if isinstance(value,self.relmodel):
-            return value.id
+    def to_python(self, value, backend=None):
+        if isinstance(value, self.relmodel):
+            return value.pkvalue()
+        elif value:
+            return self.relmodel._meta.pk.to_python(value, backend)
         else:
-            return self.relmodel._meta.pk_to_python(value)
+            return value
     json_serialize = to_python
 
     def filter(self, session, name, value):
-        fname = name.split('__')[0]
+        fname = name.split(JSPLITTER)[0]
         if fname in self.relmodel._meta.dfields:
-            return session.query(self.relmodel, fargs = {name: value})
+            return session.query(self.relmodel, fargs={name: value})
 
     def get_sorting(self, name, errorClass):
         return self.relmodel._meta.get_sorting(name, errorClass)
+    
+    def get_lookup(self, name, errorClass=ValueError):
+        if name:
+            bits = name.split(JSPLITTER)
+            fname = bits.pop(0)
+            field = self.relmodel._meta.dfields.get(fname)
+            meta = self.relmodel._meta
+            if field: # it is a field
+                nested = [(self.attname, meta)]
+                remaining = JSPLITTER.join(bits)
+                name, _nested = field.get_lookup(remaining, errorClass)
+                if _nested:
+                    nested.extend(_nested)
+                return (name, nested)
+            else:
+                raise errorClass('%s not a valid field for %s' % (fname, meta))
+        else:
+            return super(ForeignKey, self).get_lookup(name, errorClass)
 
 
 class JSONField(CharField):
@@ -654,8 +716,7 @@ behaviour and how the field is stored in the back-end server.
 
     And::
 
-        >>> m = MyModel(name = 'bla',
-                        data = {'pv': {'': 0.5, 'mean': 1, 'std': 3.5}})
+        >>> m = MyModel(name='bla', data={'pv': {'': 0.5, 'mean': 1, 'std': 3.5}})
         >>> m.cleaned_data
         {'name': 'bla', 'data__pv': 0.5, 'data__pv__mean': '1',\
  'data__pv__std': '3.5', 'data': '""'}
@@ -674,19 +735,19 @@ behaviour and how the field is stored in the back-end server.
 '''
     type = 'json object'
     internal_type = 'serialized'
-    default = None
+    _default = {}
 
     def get_encoder(self, params):
         self.as_string = params.pop('as_string',True)
-        if not self.as_string:
-            self.default = self.default if isinstance(self.default,dict) else {}
+        if not self.as_string and not isinstance(self._default, dict):
+            self._default = {}
         return encoders.Json(
                 charset = self.charset,
-                json_encoder = params.pop('encoder_class',DefaultJSONEncoder),
-                object_hook = params.pop('decoder_hook',DefaultJSONHook))
+                json_encoder = params.pop('encoder_class', DefaultJSONEncoder),
+                object_hook = params.pop('decoder_hook', DefaultJSONHook))
 
     @field_value_error
-    def to_python(self, value):
+    def to_python(self, value, backend=None):
         if value is None:
             return self.get_default()
         try:
@@ -734,6 +795,26 @@ behaviour and how the field is stored in the back-end server.
 
     def get_sorting(self, name, errorClass):
         pass
+    
+    def get_lookup(self, name, errorClass):
+        if self.as_string:
+            return super(JSONField, self).get_lookup(name, errorClass)
+        else:
+            if name:
+                name = JSPLITTER.join((self.attname, name))
+            return (name, None)
+    
+    def get_value(self, instance, *bits):
+        value = getattr(instance, self.name)
+        try:
+            for bit in bits:
+                value = value[bit]
+            if isinstance(value, dict) and '' in value:
+                value = value['']
+            return value
+        except Exception:
+            raise AttributeError
+            
 
 
 class ModelField(SymbolField):
@@ -748,7 +829,7 @@ registered in the model hash table, it can be used.'''
         return self.index_value(value)
 
     @field_value_error
-    def to_python(self, value):
+    def to_python(self, value, backend=None):
         if value and not hasattr(value,'_meta'):
             value = self.encoder.loads(value)
             return get_model_from_hash(value)
@@ -772,47 +853,46 @@ registered in the model hash table, it can be used.'''
 
 
 class ManyToManyField(Field):
-    '''A many-to-many relationship. Like :class:`ForeignKey`, it accepts
-**related_name** as extra argument.
+    '''A :ref:`many-to-many <many-to-many>` relationship.
+Like :class:`ForeignKey`, it requires a positional argument, the class
+to which the model is related and it accepts **related_name** as extra
+argument.
 
 .. attribute:: related_name
 
     Optional name to use for the relation from the related object
-    back to ``self``.
+    back to ``self``. For example::
+
+        class Group(odm.StdModel):
+            name = odm.SymbolField(unique=True)
+    
+        class User(odm.StdModel):
+            name = odm.SymbolField(unique=True)
+            groups = odm.ManyToManyField(Group, related_name='users')
+    
+    To use it::
+
+        >>> g = Group(name='developers').save()
+        >>> g.users.add(User(name='john').save())
+        >>> u.users.add(User(name='mark').save())
+    
+    and to remove::
+    
+        >>> u.following.remove(User.objects.get(name='john'))
 
 .. attribute:: through
 
-    Optional :class:`StdModel` to use for creating the many-to-many
-    relationship.
-
-For example::
-
-    class Group(odm.StdModel):
-        name = odm.SymbolField(unique = True)
-
-    class User(odm.StdModel):
-        name = odm.SymbolField(unique = True)
-        groups = odm.ManyToManyField(Group, related_name = 'users')
-
-To use it::
-
-    >>> g = Group(name = 'developers').save()
-    >>> g.users.add(User(name = 'john').save())
-    >>> u.users.add(User(name = 'mark').save())
-
-and to remove::
-
-    >>> u.following.remove(User.objects.get(name = 'john))
-
-
-Under the hood, a :class:`ManyToManyField` creates a new model *model_name*.
-If not provided, the the name will be constructed from the field name
-and the model holding the field. In the example above it would be
-*group_user*.
-This model contains two :class:`ForeignKeys`, one to model holding the
-:class:`ManyToManyField` and the other to the *related_model*.
+    An optional :class:`StdModel` to use for creating the many-to-many
+    relationship can be passed to the constructor, via the **through** keyword.
+    If such a model is not passed, under the hood, a :class:`ManyToManyField`
+    creates a new *model* with name constructed from the field name
+    and the model holding the field. In the example above it would be
+    *group_user*.
+    This model contains two :class:`ForeignKeys`, one to model holding the
+    :class:`ManyToManyField` and the other to the *related_model*.
+    
 '''
-    def __init__(self, model, through = None, related_name = None, **kwargs):
+    def __init__(self, model, through=None, related_name=None, **kwargs):
         self.through = through
         self.relmodel = model
         self.related_name = related_name
@@ -837,17 +917,51 @@ This model contains two :class:`ForeignKeys`, one to model holding the
 
     def add_to_fields(self):
         #A many to many field is a dummy field. All it does it provides a proxy
-        #for the through model.
+        #for the through model. Remove it from the fields dictionary
+        #and addit to the list of many_to_many
         self.meta.dfields.pop(self.name)
+        self.meta.manytomany.append(self.name)
 
 
-class CompositeIdField(SymbolField):
+class CompositeIdField(AutoIdField):
     '''This field can be used when an instance of a model is uniquely
 identified by a combination of two or more :class:`Field` in the model
-itself.'''
+itself. It requires a number of positional arguments greater or equal 2.
+These arguments must be fields names in the model where the
+:class:`CompositeIdField` is defined.
+
+.. attribute:: fields
+
+    list of :class:`Field` names which are used to uniquely identify a
+    model instance
+    
+Check the :ref:`composite id tutorial <tutorial-compositeid>` for more
+information and tips on how to use it.
+'''
     type = 'composite'
     def __init__(self, *fields, **kwargs):
-        kwargs['primary_key'] = True
-        super(CompositeIdField,self).__init__(**kwargs)
+        super(CompositeIdField, self).__init__(**kwargs)
         self.fields = fields
+        if len(self.fields) < 2:
+            raise FieldError('At least tow fields are required by composite '\
+                             'CompositeIdField')
 
+    def get_value(self, instance, *bits):
+        if bits:
+            raise AttributeError
+        values = tuple((getattr(instance, f.attname) for f in self.fields))
+        return hash(values)
+    
+    def register_with_model(self, name, model):
+        fields = []
+        for field in self.fields:
+            if field not in model._meta.dfields:
+                raise FieldError('Composite id field "%s" in in "%s" model.' %\
+                                 (field, model._meta))
+            field = model._meta.dfields[field]
+            if field.internal_type not in ('text', 'numeric'):
+                raise FieldError('Composite id field "%s" not valid type.' %\
+                                 field)
+            fields.append(field)
+        self.fields = tuple(fields)
+        return super(CompositeIdField, self).register_with_model(name, model)
\ No newline at end of file
