@@ -1,14 +1,15 @@
 from copy import copy
 from inspect import isgenerator
+from functools import partial
 
-from stdnet import on_result, BackendRequest, range_lookups, lookup_value
-from stdnet.exceptions import *
-from stdnet.utils import zip, JSPLITTER, iteritems, unique_tuple
+from stdnet import range_lookups
+from stdnet.utils import JSPLITTER, iteritems, unique_tuple
+from stdnet.utils.exceptions import *
 
-from .signals import *
+from .globals import lookup_value
 
 
-__all__ = ['Q', 'Query', 'QueryElement', 'EmptyQuery',
+__all__ = ['Q', 'QueryBase', 'Query', 'QueryElement', 'EmptyQuery',
            'intersect', 'union', 'difference']
 
 def iterable(value):
@@ -31,20 +32,12 @@ def update_dictionary(result, extra):
         result[k] = v
     return result
 
-
-class SessionModelBase(object):
-
-    def __init__(self, meta, session):
-        self.meta = meta
-        self.session = session
-
-    def __repr__(self):
-        return self.meta.__repr__()
-    __str__ = __repr__
-
-    @property
-    def backend(self):
-        return self.session.backend
+def get_lookups(attname, field_lookups):
+    lookups = field_lookups.get(attname)
+    if lookups is None:
+        lookups = []
+        field_lookups[attname] = lookups
+    return lookups
 
 
 class Q(object):
@@ -57,12 +50,16 @@ class Q(object):
 .. attribute:: model
 
     the :class:`StdModel` class for this query.
+    
+.. attribute:: backend
+
+    the :class:`stdnet.BackendDataServer` class for this query.
 '''
     keyword = ''
     name = ''
-    def __init__(self, meta, session, select_related = None,
-                 ordering = None, fields = None,
-                 get_field = None, name = None, keyword = None):
+    def __init__(self, meta, session, select_related=None,
+                 ordering=None, fields=None,
+                 get_field=None, name=None, keyword=None):
         self._meta = meta
         self.session = session
         self.data = {'select_related': select_related,
@@ -98,7 +95,7 @@ class Q(object):
 
     @property
     def backend(self):
-        return self.session.backend
+        return self.session.model(self._meta).read_backend
 
     def get_field(self, field):
         '''A :class:`Q` performs a series of operations and ultimately
@@ -138,10 +135,10 @@ fields.
         pass
 
     def backend_query(self, **kwargs):
-        '''Build the :class:`stdnet.BackendQuery` for this instance. This
-is a virtual method with different implementation in :class:`Query`
+        '''Build the :class:`stdnet.utils.async.BackendQuery` for this instance.
+This is a virtual method with different implementation in :class:`Query`
 and :class:`QueryElement`.'''
-        raise NotImplementedError()
+        raise NotImplementedError
 
     def _clone(self):
         cls = self.__class__
@@ -231,8 +228,8 @@ class Select(QueryElement):
 def make_select(keyword,queries):
     first = queries[0]
     queries = [q.construct() for q in queries]
-    return Select(first.meta, first.session, keyword = keyword,
-                  underlying = queries)
+    return Select(first.meta, first.session, keyword=keyword,
+                  underlying=queries)
 
 def intersect(queries):
     return make_select('intersect',queries)
@@ -250,11 +247,7 @@ def queryset(qs, **kwargs):
 class QueryBase(Q):
 
     def __iter__(self):
-        result = self.items()
-        if isinstance(result, BackendRequest):
-            raise RuntimeError('Cannot iterate. Asynchronous result not ready.')
-        else:
-            return iter(result)
+        return iter(self.items())
 
     def __len__(self):
         return self.count()
@@ -361,18 +354,19 @@ criteria and options associated with it.
 '''
     start = None
     stop = None
-    lookups = ('in','contains')
+    lookups = ('in', 'contains')
 
     def __init__(self, *args, **kwargs):
         '''A :class:`Query` is not initialized directly but via the
 :meth:`Session.query` or :meth:`Manager.query` methods.'''
-        self.fargs  = kwargs.pop('fargs',None)
-        self.eargs  = kwargs.pop('eargs',None)
-        self.unions = kwargs.pop('unions',())
-        self.intersections = kwargs.pop('intersections',())
-        self.text  = kwargs.pop('text',None)
-        self.exclude_fields = kwargs.pop('exclude_fields',None)
-        super(Query,self).__init__(*args,**kwargs)
+        self.fargs  = kwargs.pop('fargs', None)
+        self.eargs  = kwargs.pop('eargs', None)
+        self.unions = kwargs.pop('unions', ())
+        self.searchengine = kwargs.pop('searchengine', None)
+        self.intersections = kwargs.pop('intersections', ())
+        self.text  = kwargs.pop('text', None)
+        self.exclude_fields = kwargs.pop('exclude_fields', None)
+        super(Query,self).__init__(*args, **kwargs)
         self.clear()
 
     @property
@@ -383,23 +377,18 @@ criteria and options associated with it.
             return False
 
     def __repr__(self):
-        seq = self.cache().get(None)
+        seq = self.backend_query().cache.get(None) if self.executed else None
         if seq is None:
             s = self.__class__.__name__
             if self.fargs:
-                s = '%s.filter(%s)' % (s,self.fargs)
+                s = '%s.filter(%s)' % (s, self.fargs)
             if self.eargs:
-                s = '%s.exclude(%s)' % (s,self.eargs)
+                s = '%s.exclude(%s)' % (s, self.eargs)
             return s
         else:
-            return str(seq)
+            return repr(seq)
     __str__ = __repr__
 
-    def __getitem__(self, slic):
-        if isinstance(slic, slice):
-            return self.items(slic)
-        return self.items()[slic]
-
     def filter(self, **kwargs):
         '''Create a new :class:`Query` with additional clauses corresponding to
 ``where`` or ``limit`` in a ``SQL SELECT`` statement.
@@ -479,38 +468,59 @@ the :meth:`union` method.'''
         q.data['ordering'] = ordering
         return q
 
-    def search(self, text, lookup = None):
+    def search(self, text, lookup=None):
         '''Search *text* in model. A search engine needs to be installed
 for this function to be available.
 
 :parameter text: a string to search.
 :return type: a new :class:`Query` instance.
+''' 
+        q = self._clone()
+        q.text = (text, lookup)
+        return q
+
+    def where(self, code, load_only=None):
+        '''For :ref:`backend <db-index>` supporting scripting, it is possible
+to construct complex queries which execute the scripting *code* against
+each element in the query. The *coe* should reference an instance of
+:attr:`model` by ``this`` keyword.
+
+:parameter code: a valid expression in the scripting language of the database.
+:parameter load_only: Load only the selected fields when performing the query
+    (this is different from the :meth:`load_only` method which is used when
+    fetching data from the database). This field is an optimization which is
+    used by the :ref:`redis backend <redis-server>` only and can be safely
+    ignored in most use-cases.
+:return: a new :class:`Query`
 '''
-        if self._meta.searchengine:
+        if code:
             q = self._clone()
-            q.text = (text,lookup)
+            q.data['where'] = (code, load_only)
             return q
         else:
-            raise QuerySetError('Search not implemented for {0} model'\
-                                .format(self.model))
-
+            return self            
+        
     def search_queries(self, q):
         '''Return a new :class:`QueryElem` for *q* applying a text search.'''
         if self.text:
-            return self._meta.searchengine.search_model(q, *self.text)
+            searchengine = self.session.router.search_engine
+            if searchengine:
+                return searchengine.search_model(q, *self.text)
+            else:
+                raise QuerySetError('Search not available for %s' % self._meta)
         else:
             return q
 
     def load_related(self, related, *related_fields):
         '''It returns a new :class:`Query` that automatically
-follows the foreign-key relationship *related*.
+follows the foreign-key relationship ``related``.
 
 :parameter related: A field name corresponding to a :class:`ForeignKey`
     in :attr:`Query.model`.
-:parameter fields: optional :class:`Field` names for the *related*
+:parameter related_fields: optional :class:`Field` names for the ``related``
     model to load. If not provided, all fields will be loaded.
 
-This function is :ref:`performance boost <increase-performance>` when
+This function is :ref:`performance boost <performance-loadrelated>` when
 accessing the related fields of all (most) objects in your query.
 
 If Your model contains more than one foreign key, you can use this function
@@ -519,30 +529,12 @@ in a generative way::
     qs = myquery.load_related('rel1').load_related('rel2','field1','field2')
 
 :rtype: a new :class:`Query`.'''
-        meta = self._meta
-        if related in meta.dfields:
-            field = meta.dfields[related]
-            if not hasattr(field,'relmodel'):
-                raise FieldError('Load related does not apply to "{0}"'\
-                                 .format(related))
-        else:
-            raise FieldError('Unknown field "{0}"'.format(related))
+        field = self._get_related_field(related)
+        if not field:
+            raise FieldError('"%s" is not a related field for "%s"' %\
+                              (related, self._meta))
         q = self._clone()
-        rf = set(related_fields)
-        # if is always loaded.
-        rf.discard('id')
-        # we need to copy the related dictionary including its values
-        if q.select_related:
-            d = dict(((k,set(v)) for k,v in q.select_related.items()))
-        else:
-            d = {}
-        q.data['select_related'] = d
-        if field.name in d:
-            fields = d[field.name]
-            fields.update(rf)
-        else:
-            d[field.name] = rf
-        return q
+        return q._add_to_load_related(field, *related_fields)
 
     def load_only(self, *fields):
         '''This is provides a :ref:`performance boost <increase-performance>`
@@ -553,7 +545,20 @@ to the database. However, it can save you lots of bandwidth when excluding
 data intensive fields you don't need.
 '''
         q = self._clone()
-        fs = unique_tuple(q.fields, fields)
+        new_fields = []
+        for field in fields:
+            if JSPLITTER in field:
+                bits = field.split(JSPLITTER)
+                related = self._get_related_field(bits[0])
+                if related:
+                    q._add_to_load_related(related, JSPLITTER.join(bits[1:]))
+                    continue
+            new_fields.append(field)
+        if fields and not new_fields:
+            # if we added a field to the load_related list and not fields are
+            # are left we add the primary key so that other firls are not loaded.
+            new_fields.append(self._meta.pkname())
+        fs = unique_tuple(q.fields, new_fields)
         q.data['fields'] = fs if fs else None
         return q
 
@@ -567,19 +572,22 @@ to load all fields except a subset specified by *fields*.
         q.exclude_fields = fs if fs else None
         return q
 
+    ##        METHODS FOR RETRIEVING DATA
+    
+    def __getitem__(self, slic):
+        return self.backend_query()[slic]
+    
+    def items(self, callback=None):
+        '''Retrieve all items for this :class:`Query`.'''
+        return self.backend_query().items(callback=callback)
+    
     def get(self, **kwargs):
         '''Return an instance of a model matching the query. A special case is
 the query on ``id`` which provides a direct access to the :attr:`session`
 instances. If the given primary key is present in the session, the object
 is returned directly without performing any query.'''
-        id = kwargs.get('id')
-        if id is not None and len(kwargs) == 1:
-            # check the current session first
-            el = self.session.get(self.model, id)
-            if el is not None:
-                return el
-        # not there, perform the database query
-        return on_result(self.filter(**kwargs).items(), self._get)
+        return self.filter(**kwargs).items(
+                                    callback=self.model.get_unique_instance)
 
     def count(self):
         '''Return the number of objects in ``self``.
@@ -592,10 +600,7 @@ objects on the server side.'''
     def delete(self):
         '''Delete all matched elements of the :class:`Query`. It returns the
 list of ids deleted.'''
-        session = self.session
-        with session.begin() as t:
-            session.delete(self)
-        return t.deleted.get(self._meta)
+        return self.session.delete(self)
 
     def construct(self):
         '''Build the :class:`QueryElement` representing this query.'''
@@ -604,13 +609,13 @@ list of ids deleted.'''
         return self.__construct
 
     def backend_query(self, **kwargs):
-        '''Build and return the :class:`stdnet.BackendQuery`.
+        '''Build and return the :class:`stdnet.utils.async.BackendQuery`.
 This is a lazy method in the sense that it is evaluated once only and its
 result stored for future retrieval.'''
         q = self.construct()
         return q if isinstance(q, EmptyQuery) else q.backend_query(**kwargs)
 
-    def test_unique(self, fieldname, value, instance = None, exception = None):
+    def test_unique(self, fieldname, value, instance=None, exception=None):
         '''Test if a given field *fieldname* has a unique *value*
 in :attr:`model`. The field must be an index of the model.
 If the field value is not unique and the *instance* is not the same
@@ -623,33 +628,23 @@ an exception is raised.
     Default: :attr:`ModelMixin.DoesNotValidate`.
 :return: *value*
 '''
-        try:
-            r = self.get(**{fieldname:value})
-        except self.model.DoesNotExist:
-            return value
-
-        if instance and r.id == instance.id:
-            return value
-        else:
-            exception = exception or self.model.DoesNotValidate
-            raise exception('An instance with {0} {1} is already available'\
-                            .format(fieldname,value))
+        qs = self.filter(**{fieldname:value})
+        callback = partial(self._test_unique, fieldname, value,
+                           instance, exception)
+        return qs.backend_query().items(callback=callback)
 
+    def map_reduce(self, map_script, reduce_script, **kwargs):
+        '''Perform a map/reduce operation on this query.'''
+        pass
+        
     ############################################################################
     # PRIVATE METHODS
     ############################################################################
     def clear(self):
         self.__construct = None
-        self.__slice_cache = None
-
-    def cache(self):
-        if not self.__slice_cache:
-            self.__slice_cache = {}
-        return self.__slice_cache
 
     def _construct(self):
         if self.fargs:
-            args = []
             fargs = self.aggregate(self.fargs)
             for f in fargs:
                 # no values to filter on. empty result.
@@ -703,81 +698,67 @@ an exception is raised.
  Field "{1}" does not exist.'.format(meta, field_name))
             field = fields[field_name]
             attname = field.attname
+            lookup = None
             if bits:
                 bits = [n.lower() for n in bits]
                 if bits[-1] == 'in':
                     bits.pop()
-                lookup = JSPLITTER.join(bits)
-                if lookup and lookup not in range_lookups:
-                    lvalue = field.filter(self.session, lookup, value)
-                    if lvalue is not None:
-                        value = lvalue
-                        lookup = None
-                    else:
-                        if bits[-1] in range_lookups:
-                            lookup = bits.pop()
-                        else:
-                            lookup = None
-                        bits.insert(0, attname)
-                        attname = JSPLITTER.join(bits)
-            else:
-                lookup = None
-            # Get lookups on attribute name
-            lookups = field_lookups.get(attname)
-            if lookups is None:
-                lookups = []
-                field_lookups[attname] = lookups
-            if lookup not in range_lookups:
-                if not field.index:
-                    raise QuerySetError("{0} {1} is not an index.\
- Cannot query.".format(field.__class__.__name__,field_name))
-                if not iterable(value):
-                    value = (value,)
-                for v in value:
-                    if isinstance(v, Q):
-                        v = lookup_value('set', v.construct())
-                    else:
-                        v = lookup_value('value', field.dumps(v, lookup))
-                    lookups.append(v)
-            else:
-                lookups.append(lookup_value(lookup, field.dumps(value, lookup)))
+                elif bits[-1] in range_lookups:
+                    lookup = bits.pop()
+                remaining = JSPLITTER.join(bits)
+                if lookup:  # this is a range lookup
+                    attname, nested = field.get_lookup(remaining, QuerySetError)
+                    lookups = get_lookups(attname, field_lookups)
+                    lookups.append(lookup_value(lookup, (value, nested)))
+                    continue
+                elif remaining:   # Not a range lookup, must be a nested filter
+                    value = field.filter(self.session, remaining, value)
+            lookups = get_lookups(attname, field_lookups)
+            # If we are here the field must be an index
+            if not field.index:
+                raise QuerySetError("%s %s is not an index. Cannot query." %\
+                                    (field.__class__.__name__, field_name))
+            if not iterable(value):
+                value = (value,)
+            for v in value:
+                if isinstance(v, Q):
+                    v = lookup_value('set', v.construct())
+                else:
+                    v = lookup_value('value', field.dumps(v, lookup))
+                lookups.append(v)
+        #
         return [queryset(self, name=name, underlying=field_lookups[name])\
                 for name in sorted(field_lookups)]
-
-    def items(self, slic=None):
-        '''Fetch data matching theis :class:`Query` and return a list
-of instances of models.'''
-        key = None
-        seq = self.cache().get(None)
-        if slic:
-            if seq is not None:
-                return seq[slic]
-            else:
-                key = (slic.start, slic.step, slic.stop)
-        if seq is not None:
-            return seq
-        else:
-            return on_result(self.backend_query().items(slic), self._items, key)
         
-    def _items(self, items, key):
-        if isinstance(items, Exception):
-            raise items
-        session = self.session
-        seq = []
-        model = self.model
-        for el in items:
-            if isinstance(el, model):
-                session.add(el, modified=False)
-            seq.append(el)
-        self.cache()[key] = seq
-        return seq
-
-    def _get(self, items):
+    def _test_unique(self, fieldname, value, instance, exception, items):
         if items:
-            if len(items) == 1:
-                return items[0]
+            r = self.model.get_unique_instance(items)
+            if instance and r.id == instance.id:
+                return value
             else:
-                raise QuerySetError('Get query {0} yielded non\
- unique results'.format(self))
+                exception = exception or self.model.DoesNotValidate
+                raise exception('An instance with %s %s is already available'\
+                                % (fieldname, value))
         else:
-            raise self.model.DoesNotExist()
+            return value
+
+    def _get_related_field(self, related):
+        meta = self._meta
+        if related in meta.dfields:
+            field = meta.dfields[related]
+            if hasattr(field, 'relmodel'):
+                return field
+            
+    def _add_to_load_related(self, field, *related_fields):
+        rf = unique_tuple((v for v in related_fields))
+        # we need to copy the related dictionary including its values
+        if self.select_related:
+            d = dict(((k, tuple(v)) for k, v in self.select_related.items()))
+        else:
+            d = {}
+        self.data['select_related'] = d
+        if field.name in d:
+            d[field.name] = unique_tuple(d[field.name], rf)
+        else:
+            d[field.name] = rf
+        return self
\ No newline at end of file
